/** 
 *   Site testing for SKA candidate.
 *
 *   (c) Copyright 2004 Federico Bareilles <fede@iar.unlp.edu.ar>,
 *   Instituto Argentino de Radio Astronomia (IAR).
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *     
 *   The author does NOT admit liability nor provide warranty for any
 *   of this software. This material is provided "AS-IS" in the hope
 *   that it may be useful for others.
 *
 **/


#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include <sys/file.h>
#include <sys/stat.h>

#include "my_stdio.h"
#include "observ_table.h"
#include "analyzer.h"



int is_mode(char mode)
{
	int res = -1;

	switch(mode) {
	case '0':
	case 'c':
	case 'C': res = 0; break;
	case '1': res = 1; break;
	case '2': res = 2; break;
	default:  SYSLOG(LOG_ERR,"Unknow mode %c, ignored.", mode);
	}

	return res;
}

struct modes *make_modes(void)
{
	struct modes *obs;
	
	obs = (struct modes *) malloc(sizeof(struct modes));
	if( obs == NULL ){
		PVERB(PERR,"make_modes: error alocating struct modes.\n");
		return NULL;
	}
	obs->next = (struct modes *) NULL;
	obs->mode = -1;

	return obs;
}


struct sequence *make_sequence(void)
{
	struct sequence *seq;
	
	seq = (struct sequence *) malloc(sizeof(struct sequence));
	if( seq == NULL ){
		PVERB(PERR,"make_sequence: error alocating struct sequence.\n");
		return NULL;
	}
	seq->next = (struct sequence *) NULL;
	seq->mode = -1;
	seq->restart = 0;

	return seq;
}



static int get_modes(char *s, struct modes **o)
{
	int i, res;
	char mode;
	struct modes *op = *o;
	struct modes *of = *o;
	int remainder;
	double delta;
	double start_freq, stop_freq , final_freq;
	int integer;
	

	res = sscanf(s,"%c %6lf %lf %d %d %d %c",
		     &mode, &op->start_freq, &op->stop_freq, 
		     &op->RBW, &op->sweep, &op->repetitions, &op->switchs);
	if(res != 7){
		SYSLOG(LOG_ERR,
		       "get_modes: Error in modesation file (get %d)",
		      res);
		return -1;
	}      
	op->mode = is_mode(mode);
	if (op->mode < 0) return -1;
	integer = (int) ((double) op->sweep / SWEEP_MAX);
	if( integer ) {
		remainder = (int) fmod((double) op->sweep, SWEEP_MAX);
		delta = (op->stop_freq - op->start_freq ) /
			((double) integer + (double) (remainder ? 1 : 0));
		final_freq = op->stop_freq;
		op->stop_freq = op->start_freq + delta;
		op->sweep = SWEEP_MAX;
		start_freq = op->stop_freq;
		for(i=2; i <= integer  ; i++){
			op->next = make_modes(); 
			op = op->next;
			stop_freq = of->start_freq + ((double) i * delta);
			op->mode = of->mode;
			op->start_freq = start_freq;
			op->stop_freq = stop_freq;
			op->RBW = of->RBW;
			op->sweep = SWEEP_MAX;
			op->repetitions = of->repetitions;
			op->switchs = of->switchs;
			start_freq = stop_freq;
		}
		if( remainder ){
			op->next = make_modes();
			op = op->next;
			op->mode = of->mode;
			op->start_freq = start_freq;
			op->stop_freq = final_freq;
			op->RBW = of->RBW;
			op->sweep = remainder;
			op->repetitions = of->repetitions;
			op->switchs = of->switchs;
		}
	}
	*o = op;

	return 0;
}


struct modes *get_modes_table(char *file_name)
{
	FILE *fp;
	char line[160];
	int start = 0;
	int i;
	struct modes *obs, *obsp,
		*obsl = NULL;

	obsp = obs = make_modes();
	fp = fopen(file_name, "r");
	if( fp == NULL ){
		SYSLOG(LOG_ERR,
		       "get_modes_table: Error opening file: %s",file_name);
		free(obsp);
		return NULL;
	}
	while( fgets(line, sizeof(line), fp)){
		if( memcmp(line,"observ",6) == 0 && start){
			for( i=0; (i < 160) && (line[i] != '=');i++);
			i++;
			if( get_modes(stripwhite(line+i), &obsp) == 0 ){
				obsl = obsp;
				if(obsp == NULL) {
					PVERB(PERR,
				          "get_modes_table: obsp is NULL\n");
					exit(1);
				}
				obsp->next = make_modes();
				//obsp->next = (struct modes *)
                                //        malloc(sizeof(struct modes));
				obsp = obsp->next;
				obsp->next = NULL;
			}
		}else{
			if( memcmp(line,"start",5) == 0 ) start = 1;
			if( memcmp(line,"stop",4) == 0 )  start = 0;
		}
	}
	fclose(fp);

	if ( obsl ) {
		free_modes_table(obsp);
		obsl->next = NULL;
	}

	return obs;
}


int free_modes_table(struct modes *obs)
{
	struct modes *op = obs;
	struct modes *opp;

	while(op != NULL){
		opp = op;
		op = op->next;
		free(opp);
	}

	return 0;
}


int free_sequence_table(struct sequence *seq)
{
	struct sequence *sp = seq;
	struct sequence *spp;

	while(sp != NULL){
		spp = sp;
		sp = sp->next;
		free(spp);
	}

	return 0;
}

static int get_sequence(char *s, struct sequence **o)
{
	int res;
	char mode;
	struct sequence *sp = *o;

	res = sscanf(s,"%c %d %c",
		     &mode, &sp->azimuth, &sp->polarization);
	if(res != 3){
		SYSLOG(LOG_ERR,
		       "get_sequence: Error in sequence file (get %d)", res);
		return -1;
	}      
	sp->mode = is_mode(mode);
	if (sp->mode < 0) return -1;

	return 0;
}



struct sequence *get_sequence_table(char *file_name)
{
	FILE *fp;
	char line[160];
	int start = 0;
	int i;
	struct sequence *seq, *seqp, 
		*seql = NULL;

	seqp = seq = make_sequence();
	fp = fopen(file_name, "r");
	if( fp == NULL ){
		SYSLOG(LOG_ERR,
		       "get_sequence_table: Error opening file: %s",
		       file_name);
		free(seqp);
		return NULL;
	}
	while( fgets(line, sizeof(line), fp)){
		if( memcmp(line,"seq",3) == 0 && start){
			for( i=0; (i < 160) && (line[i] != '=');i++);
			i++;
			if( get_sequence(stripwhite(line+i), &seqp) == 0 ){
				seql = seqp;
				if(seqp == NULL) {
					PVERB(PERR,
					      "get_sequence_table: obsp is NULL\n");
					exit(1);
				}
				seqp->next = make_sequence();
				//seqp->next = (struct sequence *)
				//       malloc(sizeof(struct sequence));
				seqp = seqp->next;
				seqp->next = NULL;
			}
		}else{
			if( memcmp(line,"start",5) == 0 ) start = 1;
			if( memcmp(line,"stop",4) == 0 )  start = 0;
			/* TODO */
			if( memcmp(line,"restart",7) == 0 ) {
				seql->restart = 1;
				start = 0;
			}
		}
	}
	fclose(fp);

	if ( seql ) {
		free_sequence_table(seqp);
		seql->next = NULL;
	}


	return seq;
}


time_t time_estimation_mode(int mode, struct modes *mod)
{
	struct modes *modp = mod;
	double est = 0.0;

	while (modp != NULL) {
		if (modp->mode == mode)
			est += ((double)modp->sweep + SOME_LOST_TIME) *
				(double)modp->repetitions;
		modp = modp->next;
	}
		
	return (time_t) est;
}
