/** 
 *   Site testing for SKA candidate.
 *
 *   (c) Copyright 2004 Federico Bareilles <fede@iar.unlp.edu.ar>,
 *   Instituto Argentino de Radio Astronomia (IAR).
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *     
 *   The author does NOT admit liability nor provide warranty for any
 *   of this software. This material is provided "AS-IS" in the hope
 *   that it may be useful for others.
 *
 *   TODO:
 *        *) Data save.
 *        *)
 *
 *
 **/

#define _GNU_SOURCE

//#define FITS_COMPAT

#include <stdio.h>
#include <unistd.h>
#include <time.h>
#include <sys/file.h>


#include "observ_table.h"
#include "analyzer.h"
#include "my_stdio.h"
#include "switch_rf.h"
#include "rotator.h"

#define H_PREFIX "# "


int print_observ_header(int fd, struct obs_setup *s)
{
	struct tm *t;
	
	t = localtime(&s->start);
	fd_printf(fd, H_PREFIX"NAXIS   = %d\n", 2);
	fd_printf(fd, H_PREFIX"NAXIS1  = %d\n", AN_MAX_POINTS);
	fd_printf(fd, H_PREFIX"NAXIS2  = %d\n", 1);
#ifdef FITS_COMPAT
	fd_printf(fd, H_PREFIX"CTYPE1  = 'FREQ'\n");
	fd_printf(fd, H_PREFIX"CDELT1  = %14f\n", 
		  (s->obs->stop_freq - s->obs->start_freq)/
		  ((double) (AN_MAX_POINTS - 1)) * 1e6);
	fd_printf(fd, H_PREFIX"CRVAL1  = %14f\n", 
		  s->obs->start_freq * 1e6);

	fd_printf(fd, H_PREFIX"CTYPE2  = 'db'\n");
	fd_printf(fd, H_PREFIX"CDELT2  = %14f\n", 0.0);
	fd_printf(fd, H_PREFIX"CRVAL2  = %14f\n", 1.0);
#endif
	fd_printf(fd, H_PREFIX"DATE = %04d-%02d-%02d\n",
		  t->tm_year + 1900, t->tm_mon + 1, t->tm_mday);
	fd_printf(fd, H_PREFIX"TIME = %02d:%02d:%02d\n",
		  t->tm_hour, t->tm_min, t->tm_sec);
	fd_printf(fd, H_PREFIX"DATE-UNIX = %d\n", s->start);
	fd_printf(fd, H_PREFIX"AZIMUTH = %3d\n", s->azimuth); 

	fd_printf(fd, H_PREFIX"MODE = %d\n", s->obs->mode); 
	fd_printf(fd, H_PREFIX"START-FREQ = %8.5f\n", s->obs->start_freq); 
	fd_printf(fd, H_PREFIX"STOP-FREQ  = %8.5f\n", s->obs->stop_freq); 
	fd_printf(fd, H_PREFIX"RBW = %4d\n", s->obs->RBW); 
	fd_printf(fd, H_PREFIX"SWEEP = %4d\n", s->obs->sweep); 
	fd_printf(fd, H_PREFIX"REPETITION = %2d of %2d\n", 
		  s->current_rep, s->obs->repetitions); 
	fd_printf(fd, H_PREFIX"ANTENNA-CONF = %c\n", s->obs->switchs); 
	fd_printf(fd, H_PREFIX"POLARIZATION = %c\n", s->polarization); 
	fd_printf(fd, H_PREFIX"END\n");

	return 0;
}


int save_observ(struct obs_setup *s)
{
	char file[160];
	struct tm *t;
	int fd, i;
/* 
   s2004112617401030h00959999011000004418A
   |                                      \__________
   |                                                 \
   s 2004 11 26 17 40 1 030 h 00959999 01100000 44 18A
   | |    |  |  |  |  | |   | |        |        |  | +-> Antenna Conf. 
   | |    |  |  |  |  | |   | |        |        |  +---> Current repetition
   | |    |  |  |  |  | |   | |        |        +------> Repetitions
   | |    |  |  |  |  | |   | |        +---------------> Stop freq [Hz]
   | |    |  |  |  |  | |   | +------------------------> Start freq [Hz]
   | |    |  |  |  |  | |   +--------------------------> Polarization
   | |    |  |  |  |  | +------------------------------> Azimuth
   | |    |  |  |  |  +--------------------------------> Mode
   | |    |  |  |  +-----------------------------------> Minutes
   | |    |  |  +--------------------------------------> Hour
   | |    |  +-----------------------------------------> Day
   | |    +--------------------------------------------> Month
   | +-------------------------------------------------> Year
   +---------------------------------------------------> ID 

*/
	t = localtime(&s->start);
	sprintf(file,"%s/s%04d%02d%02d%02d%02d%1d%03d%c%08.0f%08.0f%02d%02d%c",
		s->data_dir, t->tm_year + 1900, t->tm_mon + 1, t->tm_mday,
		t->tm_hour, t->tm_min, s->mode, s->azimuth, s->polarization,
		(s->obs->start_freq * 1000000.0),
		(s->obs->stop_freq * 1000000.0),
		s->obs->repetitions, s->current_rep, s->obs->switchs);

	//sprintf(file,"s2004");

	fd = open(file, O_WRONLY|O_CREAT|O_TRUNC, 
		  S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
	if (fd < 0) {
		SYSLOG(LOG_ERR,"save_observ: Error opening file: %s", file);
		return -1;
	}
	/* print header */
	print_observ_header(fd, s);
	/* print data */
	for(i=0;i<AN_MAX_POINTS;i++)
		fd_printf(fd, "%8.2f\n", *(s->data+i));

	close(fd);

	return 0;
	
}


int get_observ( int dev, struct obs_setup *s)
{
	int ret;
	int i = 0;

	while (analyzer_get_observ( dev, s) != 0) {
		
		if ((ret = analyzer_reset(dev)) != 0){
		  if (i < AN_RETRY_TIME_OUT) {
		    sleep(1);
		    i++;
		  }
		  else
		    return ret;
		}
	}
	
	return 0;
}


/**
 * Check and set rotator position
 *
 * Return:
 *        -1: current position not equal to setup, try to seting this.
 *         0: position is equal to setup, do nothing.
 *
 **/
int chk_set_az_pol(struct obs_setup *s)
{
	int az;
	char pol;
	int ret = 0;

	rotator_get_azimuth(s->fd_rot, &az);
	rotator_get_polarization(s->fd_rot, &pol);

	if ((s->azimuth != az) || (s->polarization != pol)) {
		ret = -1;
		if (s->azimuth != az) 
			rotator_set_azimuth(s->fd_rot, s->azimuth);
		if (s->polarization != pol)
			rotator_set_polarization(s->fd_rot, s->polarization);
	}

	return ret;
}


int observ_mode(int dev, struct obs_setup *s)
{
	struct modes *obsp = s->obs;
	int i, error;
	char current_switchs = -1;
	int done;

	antenna_switchs_set(current_switchs);
	while (obsp != NULL ) {
		if (obsp->mode == s->mode) {
			if (current_switchs != obsp->switchs) {
				if (antenna_switchs_set(obsp->switchs) < 0) {
					SYSLOG(LOG_ERR, "Cant set rf switch.");
				}
			}
			for(i = 0; i < obsp->repetitions; i++){
				s->current_rep = i+1;
				s->obs = obsp;
				chk_set_az_pol(s);
				done = 0;
				do {
					error = get_observ( dev, s);
					if (error) {
						return error;
					}
					if (chk_set_az_pol(s) == 0) {
						done = 1;
					}else{
						SYSLOG(LOG_ERR, "rotator position error.");
					}
				} while(!done);
				save_observ(s);

			}
		}
		obsp = obsp->next;
	}

	return 0;
}
