/** 
 *   Site testing for SKA candidate.
 *
 *   (c) Copyright 2004 Federico Bareilles <fede@iar.unlp.edu.ar>,
 *   Instituto Argentino de Radio Astronomia (IAR).
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *     
 *   The author does NOT admit liability nor provide warranty for any
 *   of this software. This material is provided "AS-IS" in the hope
 *   that it may be useful for others.
 *
 **/
#define _GNU_SOURCE

#include <ctype.h>
#include <stdio.h>
#include <termios.h>
#include <unistd.h>
#include <sys/file.h>
#include <string.h>

#include "my_stdio.h"
#include "rotator.h"
#include "observ_table.h"

#define PK_LEN   5
/* command string definition: */
#define PK_STX   0
#define PK_CMD   1 
#define PK_DATA1 2 
#define PK_DATA0 3
#define PK_CHS   4 

#define STX_DATA    0xff  /* Start of Transmision */
#define DC_DATA     0x00  /* Don't care */

#define SET_RESET   0x00
#define SET_AZ      0x03
#define SET_POL     0x01

#define ROT_ERR     0x02

#define GET_AZ      0x05
#define GET_POL     0x04
#define GET_ERR     ROT_ERR

#define POL_HORIZONTAL  0
#define POL_VERTICAL    1

#define TO_ROTATOR_L(v) (v & 0xff)
#define TO_ROTATOR_H(v) ((v >> 8) & 0x3f)
#define FROM_ROTATOR(pk) (((pk[PK_DATA1] << 8) | pk[PK_DATA0]) & 0x3fff)

#define ROT_AZ_CONST (25.0/9.0)
#define DEG2ROT(a) ((int)((long) a * ROT_AZ_CONST))
#define ROT2DEG(a) ((int)((long) a / ROT_AZ_CONST))


struct termios oldtio;


int rt_check_packet(char *pk);


int rotator_setup(char *name)
{
	int fd;
	struct termios tio;

#ifdef DEMO_MODE 
	return 2;
#endif
	/* Open Serial Port : 8 bits data, 1 bit stop, No parity, 1200
           baud */
	if ((fd = open(name, O_RDWR | O_NOCTTY)) < 0 ){
		PVERB(PERR, "Cannot open serial port (%s).\n", name);
		return -1;
	}

	tcgetattr(fd,&oldtio); /* save current port settings */
	tcgetattr(fd,&tio);

	//bzero(&tio, sizeof(tio));
	
	tio.c_cflag &= ~CRTSCTS ;
	/* Zero out the baud rate portion of c_cflag */
	tio.c_cflag &= ~CBAUD;
	/* Set baud rate to 1200 */
	tio.c_cflag |= B1200;
	tio.c_cflag &= ~(CSIZE|PARENB);
	tio.c_cflag |= CS8 ;
	/* Ignore modem control lines */
	tio.c_cflag |= CLOCAL ;

	/* Turn of Canonical Processing and echo */
	tio.c_lflag &= ~ICANON;
	tio.c_lflag &= ~ECHO;
	/* Set MIN to one and TIME to zero */
	tio.c_cc[VMIN]  = 1;
	tio.c_cc[VTIME] = 0;

	/* Ignore BREAK condition on input */
	tio.c_iflag |= IGNBRK;
	/* ignore framing errors and parity errors */
	tio.c_iflag |=IGNPAR;
	/* Disable NL to CR mapping */
	tio.c_iflag &= ~ICRNL;
	/* Disable XON/XOFF flow control on output and input */
	tio.c_iflag &= ~(IXON|IXOFF);

	/* now clean the modem line and activate the settings for modem */
	tcflush(fd, TCOFLUSH|TCIFLUSH);
	tcsetattr(fd, TCSANOW, &tio);
	
	return fd;
}

int rotator_close(int fd)
{
#ifndef DEMO_MODE
	//rotator_set_azimuth(fd, ROTATOR_AZ_PARK);
	tcsetattr(fd,TCSANOW,&oldtio);
	close(fd);
#endif
	return 0;
}


int rt_log_bad_packet(void)
{
	SYSLOG(LOG_ERR, "Bad packet from rotator.");

	return -1;
}


ssize_t soft_read_time_out(struct timeval tv, int fd, void *buf, size_t count)
{
	fd_set rfds;
	int fd_max = fd + 1;
	int retval;

	FD_ZERO(&rfds);
	FD_SET(fd, &rfds);
	retval = select(fd_max, &rfds, NULL, NULL, &tv);
	if (retval) {
		if( FD_ISSET(fd, &rfds) ) {
			retval = read(fd, buf, count);
#ifdef DEBUG_MODE
			if (verbose_flag & PINFO) {
				int __i;
				PVERB(PINFO, "# rotator.c read <- ");
				for(__i=0;__i<retval;__i++)
					PVERB(PINFO, "0x%02x ", 
					      *(((u_char *)buf)+__i) );
				PVERB(PINFO, "\n");
			}
#endif
			return retval;
		}
	}

	return ROT_ERR_TIME_OUT;
}


int discard_noise(int fd)
{
	struct timeval tv;
	char noise;
	int i;
	char pk[PK_LEN];

	tv.tv_sec = 0;
	tv.tv_usec = 0;
	i = 0;
	while ( soft_read_time_out(tv, fd, &noise, 1) > 0) {
		if ( i && i < PK_LEN) {
			pk[i] = noise;
			i++;
		}
		if ( i == PK_LEN ) {
			if( rt_check_packet(pk) == 0) {
				/* is real packet, not noise */
				/* TODO */
				PVERB(PINFO, "Real packer arrived.\n");
				i = 0;
			}
		}
		if ( noise == (char) STX_DATA ) {
			pk[0] = STX_DATA;
			i = 1;
		}
	}
	
	return 0;
}


ssize_t soft_read(int fd, void *buf, size_t count)
{
	struct timeval tv;

	tv.tv_sec = ROTATOR_ECHO_TIME_OUT;
	tv.tv_usec = 0;

	return soft_read_time_out(tv, fd, buf, count);
}


ssize_t soft_write(int fd, const void *buf, size_t count)
{
#ifdef DEBUG_MODE
	int i;
	if (verbose_flag & PINFO) {
		PVERB(PINFO, "# rotator.c send -> ");
		for(i=0;i<count;i++)
			PVERB(PINFO, "0x%02x ", *(((u_char *)buf)+i) );
		PVERB(PINFO, "\n");
	}
#endif
#ifdef DEMO_MODE
	return count;
#endif
	discard_noise(fd);
	return write(fd, buf, count);

}

	
int rt_check_packet(char *pk)
{
	int i;
	char chs = pk[PK_STX];
	
	for (i=1; i < PK_LEN - 1; i++) chs ^= pk[i];
	if ((pk[PK_STX] == (char) STX_DATA) && (chs == pk[PK_CHS])) {
		return 0;
	}
	rt_log_bad_packet();

	return -1;
}


int rt_wait_confitmation(int fd, char *pk)
{
	struct timeval tv;
	char answer[PK_LEN];
	int res;

	tv.tv_sec = ROTATOR_TIME_OUT;
	tv.tv_usec = 0;	
	if( soft_read_time_out(tv, fd, answer, 5) < 0 ) {
		SYSLOG(LOG_ERR, "azimuth rotator time out.");
	}else{
		if((res = strncmp( answer, pk, PK_LEN)) != 0) {
			if (rt_check_packet( answer ) == 0) {
				/* valid packet */
				switch (answer[PK_CMD]) {
				case ROT_ERR:
					res = ROT_ERR_CONDITION;
					/* strncpy(pk, answer, PK_LEN); */
					break;
				default:
					SYSLOG(LOG_ERR,
					       "rotator: packet out of sequence 0x%02x", 
					       answer[PK_CMD]);
					res = -1;
				}
			}else{
				res = ROT_ERR_INVALID_PK;
			}
		}
		return res;
	}
	
	return ROT_ERR_TIME_OUT;
}
	

int rt_make_packet(char *pk, char cmd, char data1, char data0)
{
	int i;
	
	pk[PK_STX] = STX_DATA;
	pk[PK_CMD] = cmd;
	pk[PK_DATA1] = data1;
	pk[PK_DATA0] = data0;
	pk[PK_CHS] = pk[PK_STX];
	for (i=1; i< PK_LEN - 1; i++) 
		pk[PK_CHS] ^= pk[i];

	return 0;
}


int rt_rcv_packet(int fd, char *pk)
{
	int res;

	res = soft_read(fd, pk, PK_LEN);
	if (res == ROT_ERR_TIME_OUT) {
		SYSLOG(LOG_ERR, "rotator: read packet time out.");
	}

	return res;
}


int rt_send_packet(int fd, char *pk)
{
	char answer[PK_LEN];
	int res = -1;
	int i;

	for (i=0; i < 3 && res != 0; i++) {
		soft_write(fd, pk, PK_LEN);
		res = rt_rcv_packet(fd, answer);
		if (res == ROT_ERR_TIME_OUT) {
			break;
		}
		if((res = strncmp( answer, pk, PK_LEN)) != 0) {
			rt_log_bad_packet();
		}
	}

	return res;
}


/******************************************************************************
 * Rotator commands:
 **/


int rotator_reset(int fd)
{
	char pk[PK_LEN];
	int res;

	rt_make_packet(pk, SET_RESET, DC_DATA, DC_DATA);
	if ((res = rt_send_packet(fd, pk)) != 0) 
		return res;
	
	return rt_wait_confitmation(fd, pk);
}


int rotator_set_azimuth(int fd, int az)
{
	char pk[PK_LEN];
	int res;

	az = DEG2ROT(az);
	rt_make_packet(pk, SET_AZ, TO_ROTATOR_H(az), TO_ROTATOR_L(az));
	if ((res = rt_send_packet(fd, pk)) != 0) 
		return res;

	return rt_wait_confitmation(fd, pk);
}


int rotator_set_polarization(int fd, char pol)
{
	char pk[PK_LEN];
	int res;
	char poln;
	
	if (pol == HORIZONTAL) poln = POL_HORIZONTAL;
	else if (pol == VERTICAL) poln = POL_VERTICAL;
	else return -1;
	
	rt_make_packet(pk, SET_POL, DC_DATA, TO_ROTATOR_L(poln));
	if ((res = rt_send_packet(fd, pk)) != 0) 
		return res;

	return rt_wait_confitmation(fd, pk);
}


int rotator_get_azimuth(int fd, int *az)
{
	char pk[PK_LEN];
	int res;

	rt_make_packet(pk, GET_AZ, DC_DATA, DC_DATA);
	if ((res = rt_send_packet(fd, pk)) != 0) 
		return res;
	if ( rt_rcv_packet(fd, pk) != 0 ){
		return ROT_ERR_TIME_OUT;
	}
	if ((pk[PK_CMD] != (char) GET_AZ) || rt_check_packet(pk) != 0) {
		return ROT_ERR_INVALID_PK;
	}
	*az = ROT2DEG(FROM_ROTATOR(pk));

	return 0;
}


int rotator_get_polarization(int fd, char *pol)
{
	char pk[PK_LEN];
	int res;

	rt_make_packet(pk, GET_POL, DC_DATA, DC_DATA);
	if ((res = rt_send_packet(fd, pk)) != 0) 
		return res;
	if ( rt_rcv_packet(fd, pk) != 0 ){
		return ROT_ERR_TIME_OUT;
	}
	if ((pk[PK_CMD] != (char) GET_POL) || rt_check_packet(pk) != 0) {
		return ROT_ERR_INVALID_PK;
	}
	if(pk[PK_DATA0] == POL_HORIZONTAL)
		*pol = HORIZONTAL;
	else
		*pol = VERTICAL;

	return 0;
}


